// Vgm2Midi - Ben Ryves 2006
using System;
using System.Collections;
using System.Text;
using System.IO;
using System.IO.Compression;

namespace Vgm2Midi {
    // VGM file reader class
    public class VgmFile {

        /// <summary>
        /// Class to hold Japanese/English string pair
        /// </summary>
        public class MultilingualString {

            public string English = "";
            public string Japanese = "";

            public override string ToString() {
                return English;
                // Snipped, I didn't see much need for multilingual support.
                /*string Return = Program.FavourEnglish ? English : Japanese;
                if (Return == "") {
                    Return = Program.FavourEnglish ? Japanese : English;
                    if (Return == "") {
                        return "(Unknown)";
                    } else {
                        return Return;
                    }
                } else {
                    if (!Program.DisplayAllDetails) {
                        return Return;
                    } else if (Program.FavourEnglish && Japanese != "" && English != Japanese) {
                        return Return + " (" + Japanese + ")";
                    } else if (!Program.FavourEnglish && English != "" && Japanese != English) {
                        return Return + " (" + English + ")";
                    } else {
                        return Return;
                    }
                }*/
            }

            public MultilingualString() { }
            public MultilingualString(string English, string Japanese) {
                this.English = English;
                this.Japanese = Japanese;
            }

        }

        public string Length {
            get {
                TimeSpan TotalLength = new TimeSpan(((long)this.SampleCount * (long)10000000) / (long)44100);
                return TotalLength.Minutes.ToString() + ":" + TotalLength.Seconds.ToString("D2");
            }
        }

        public bool HasGd3Info = false;
        public MultilingualString Track = new MultilingualString();
        public MultilingualString Game = new MultilingualString();
        public MultilingualString System = new MultilingualString();
        public MultilingualString Author = new MultilingualString();
        public string DateReleased = "";
        public string Ripper = "";
        public string Notes = "";

        private uint FileEndOffset = 0;
        public decimal VgmVersionNumber = 0;
        public uint PsgSpeed = 0;
        public uint Y2413Speed = 0;
        private uint Gd3TagOffset = 0;
        public uint SampleCount = 0;
        public uint LoopOffset = 0;
        public uint LoopSamples = 0;
        private uint RecordingRate = 0;
        private uint Gd3Version = 0;
        private uint Gd3Size = 0;


        public byte[] VgmData;

        // VGM 1.10 additions
        public ushort LsfrTappedBits = 0x0009;
        public byte LsfrWidth = 16;

        // VGM 1.50 additions
        public uint DataOffset = 0x40;

        /// <summary>
        /// Initialise and read the file.
        /// </summary>
        /// <param name="filename">File to read from</param>
        public VgmFile(string filename) {
            MemoryStream DecompressedVgm = new MemoryStream();

            // Try and load the VGM file:

            byte[] ValidHeader = { (byte)'V', (byte)'g', (byte)'m', (byte)' ' };

            using (FileStream VgmFilestream = new FileStream(filename, FileMode.Open)) {
                byte[] CheckIfCompressed = new byte[4];
                VgmFilestream.Read(CheckIfCompressed, 0, 4);
                VgmFilestream.Seek(0, SeekOrigin.Begin);

                bool HasValidHeader = true;
                for (int i = 0; i < 4; ++i) {
                    if (CheckIfCompressed[i] != ValidHeader[i]) {
                        HasValidHeader = false;
                        break;
                    }
                }
                if (HasValidHeader) {
                    byte[] Data = new byte[VgmFilestream.Length];
                    VgmFilestream.Read(Data, 0, (int)VgmFilestream.Length);
                    DecompressedVgm.Write(Data, 0, (int)VgmFilestream.Length);
                } else {
                    // Assume that it's compressed.
                    using (GZipStream Deflator = new GZipStream(VgmFilestream, CompressionMode.Decompress, true)) {
                        byte[] Chunk = new byte[2048];

                        int BytesRead = Deflator.Read(Chunk, 0, 2048);

                        while (BytesRead > 0) {
                            DecompressedVgm.Write(Chunk, 0, BytesRead);
                            BytesRead = Deflator.Read(Chunk, 0, 2048);
                        }
                    }
                }
            }



            DecompressedVgm.Seek(0, SeekOrigin.Begin);
            using (BinaryReader VgmReader = new BinaryReader(DecompressedVgm)) {

                // Check the header again.
                byte[] CheckHeader = VgmReader.ReadBytes(4);
                for (int i = 0; i < 4; ++i) if (CheckHeader[i] != ValidHeader[i]) throw new Exception("File does not have a valid VGM header.");

                FileEndOffset = VgmReader.ReadUInt32();

                uint ExtractVersion = VgmReader.ReadUInt32();

                uint Minor = (ExtractVersion & 0x0F) + ((ExtractVersion & 0xF0) >> 4) * 10;
                ExtractVersion >>= 8;
                uint Major = (ExtractVersion & 0x0F) + ((ExtractVersion & 0xF0) >> 4) * 10;
                this.VgmVersionNumber = (decimal)((decimal)Major + ((decimal)Minor) / 100.0m);

                PsgSpeed = VgmReader.ReadUInt32();
                // Carry on loading:
                Y2413Speed = VgmReader.ReadUInt32();
                Gd3TagOffset = VgmReader.ReadUInt32();
                SampleCount = VgmReader.ReadUInt32();
                LoopOffset = VgmReader.ReadUInt32();
                LoopSamples = VgmReader.ReadUInt32();
                RecordingRate = VgmReader.ReadUInt32();


                // LSFR (VGM >=1.10)
                if (VgmVersionNumber >= 1.10m) {
                    VgmReader.BaseStream.Seek(0x28, SeekOrigin.Begin);
                    LsfrTappedBits = VgmReader.ReadUInt16();
                    LsfrWidth = VgmReader.ReadByte();
                }

                // Data offset
                if (VgmVersionNumber >= 1.50m) {
                    VgmReader.BaseStream.Seek(0x34, SeekOrigin.Begin);
                    DataOffset = VgmReader.ReadUInt32() + 0x34;
                }

                // At this point, do a wild jump to location 0x40 to load the actual VGM data:
                VgmReader.BaseStream.Seek(DataOffset, SeekOrigin.Begin);
                VgmData = VgmReader.ReadBytes((int)(FileEndOffset - 0x44));

                // Finally, check for any GD3 tag:
                if (Gd3TagOffset != 0) {
                    HasGd3Info = true;
                    VgmReader.BaseStream.Seek(Gd3TagOffset + 0x14, SeekOrigin.Begin);
                    // Check for the correct header:
                    char[] CheckGd3Header = VgmReader.ReadChars(4);
                    char[] ValidGd3Header = "Gd3 ".ToCharArray();
                    for (int i = 0; i < 4; ++i) {
                        if (CheckGd3Header[i] != ValidGd3Header[i]) {
                            throw new Exception("Not a valid GD3 header!");
                        }
                    }

                    // First, get the GD3 version and size data:
                    Gd3Version = VgmReader.ReadUInt32();
                    Gd3Size = VgmReader.ReadUInt32();

                    // Now let's go and read in that data!
                    int Gd3SizeCounter = (int)Gd3Size;
                    int TagType = 0;

                    while (Gd3SizeCounter > 0) {
                        string TagData = "";
                        while (true) {
                            char C = (char)(VgmReader.ReadByte() + VgmReader.ReadByte() * 256);
                            Gd3SizeCounter -= 2;
                            if (C != 0) {
                                TagData += C;
                            } else {
                                break;
                            }
                        }
                        // File the read tag data into the correct slot:
                        switch (TagType) {
                            case 0: Track.English = TagData; break;
                            case 1: Track.Japanese = TagData; break;
                            case 2: Game.English = TagData; break;
                            case 3: Game.Japanese = TagData; break;
                            case 4: System.English = TagData; break;
                            case 5: System.Japanese = TagData; break;
                            case 6: Author.English = TagData; break;
                            case 7: Author.Japanese = TagData; break;
                            case 8: DateReleased = TagData; break;
                            case 9: Ripper = TagData; break;
                            case 10: Notes = TagData; break;
                        }
                        ++TagType;
                    }
                }
            }
        }
    }
}
